import java.io.*;
import ij.*;
import ij.io.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import java.util.*;
import ij.plugin.*;
import ij.plugin.frame.*;

    /**
      This plugin opens DICOM datasets with fidelity and sorts into a Hyperstack.

      @author Fred Damen <fred@damen.org>

      Version History:
      2019-04-01: Created
    */

public class DICOM_open implements PlugIn, DialogListener {

   Point loc = null;
   Label where = null;
   MultiLineLabel info = null;
   Choice which = null;
   Choice diff  = null;
   TextField var = null;
   TextField tag = null;

   File path = new File(OpenDialog.getDefaultDirectory());
   String[] dirs = null;
   Map<String,String> tags = new TreeMap<String,String>();
   Map<String,String> vars = new TreeMap<String,String>();

   public void run(String arg) {
      if (IJ.versionLessThan("1.52i")) // after Prefs.ignoreRescaleSlope was added???
         return;

      initTagsVars();

      DirectoryChooser dc = new DirectoryChooser("Select Root Directory of your DICOM datasets");
      String sdir = dc.getDirectory();
      if (sdir == null) return;
      path = new File(sdir);

      loc = (Point)IJ.getProperty("PreferedLocation");
      GenericDialog gd = new GenericDialog("Select DICOM dataset");
      if (loc != null) gd.setLocation(loc.x, loc.y);
      gd.addMessage(path.getAbsolutePath(), new Font("Monospaced", Font.BOLD, 12));
      where = (Label)gd.getMessage();

      gd.addChoice("", scan(path),"");
      which = (Choice)gd.getChoices().get(0);
      gd.addToSameRow();
      gd.addCheckbox("Info",false);
      ((Checkbox)gd.getCheckboxes().get(0)).setEnabled(false);

      gd.addMessage(padR("Subject",50)+"\n"+
                    padR("Date",50)+"\n"+
                    padR("Description",50), new Font("Monospaced", Font.BOLD, 12)); 
      info = (MultiLineLabel)gd.getMessage();
      updateFields(path);

      gd.addChoice("Diffusion", new String[]{"<Nope>","Trace","X", "Y", "Z"},"<Nope>");
      diff = (Choice)gd.getChoices().get(1);
      gd.addToSameRow();
      gd.addStringField("Var",""); gd.addToSameRow(); gd.addStringField("Tag","",10);
      var = (TextField)gd.getStringFields().get(0);
      tag = (TextField)gd.getStringFields().get(1);

      gd.addRadioButtonGroup("Format:", new String[]{"Float","Scaled", "16-bit fixed", "16-bit broken"}, 1, 4, "16-bit fixed");

      gd.addDialogListener(this);
      gd.setOKLabel("Tell Me More");
      gd.showDialog();
      if (gd.wasCanceled())
         return;

      loc = gd.getLocation();
      IJ.setProperty("PreferedLocation",loc);
      ImageWindow.setNextLocation(loc.x, loc.y);

      File series = new File(where.getText());
      if (!isSeries(series))
         series = new File(series, which.getSelectedItem().split(": *",2)[1]);

      String format = gd.getNextRadioButton();

      if (diff.getSelectedIndex()>0) {
         try {
            ImagePlus imp = openDICOMDiff(series, format, diff.getSelectedItem());
            ImagePlus[] DWIimp = diffusionDICOM(imp);
            int ugh = diff.getSelectedIndex()-1;
            if (DWIimp[ugh] == null) return;
            DWIimp[ugh].changes = false;
            DWIimp[ugh].show();
            IJ.run(DWIimp[ugh], "Enhance Contrast", "saturated=0.35");

            double tr = getDouble(imp,1,"0018,0080")/1000.0;
            double acqtime = 0.0;
            for(int f=0; f<DWIimp[ugh].getNFrames(); f++) {
               int si = DWIimp[ugh].getStackIndex(1,1,f+1);
               int nex = getInt(DWIimp[ugh],si,"0018,0083");
               IJ.log(DWIimp[ugh].getStack().getSliceLabel(si).split("\n",2)[0]+"  "+nex);
               acqtime += nex * 3 * tr;
               }
            IJ.log("acqtime = "+IJ.d2s(acqtime/60.0,1)+" min");

            return;
            }
         catch(Exception e) {
            StringWriter sw = new StringWriter();
            e.printStackTrace(new PrintWriter(sw));
            IJ.error("An error occurred.\n\n"+sw.toString());
            }
         return;
         }

      String var = gd.getNextString();
      String tag = gd.getNextString();
      if (!tag.equals("")) {
         ImagePlus imp = openDICOM(series, format);
         ImagePlus Aimp = arrayDICOM(imp,tag, var.equals("") ? "v" : var);
         if (Aimp == null) return;
         Aimp.show();
         IJ.run(Aimp, "Enhance Contrast", "saturated=0.35");
         return;
         }

      ImagePlus imp = openDICOM(series, format); 
      imp = volumeDICOM(imp);
      if (imp == null) return;
      imp.show();
      IJ.run(imp, "Enhance Contrast", "saturated=0.35");
      }

   private void initTagsVars() {
      vars.put("0018,0080","tr");
      vars.put("0018,0081","te");
      vars.put("0018,1314","fa");
      for(Map.Entry<String,String> me : vars.entrySet())
         tags.put(me.getValue(),me.getKey());
      }

   public static String padR(String s, int n) {
      return String.format("%1$-" + n + "s", s);  
      }
   public static String padL(String s, int n) {
      return String.format("%1$" + n + "s", s);  
      }

   public static boolean isDICOM(File file) {
      if (file == null) return false;
      try {
         InputStream is = new FileInputStream(file);
         if (is.read(new byte[8*16]) == -1) 
            return false;
         byte[] fourbytes = new byte[4];
         if (is.read(fourbytes) == -1) 
            return false;
         if ((new String(fourbytes)).equals("DICM"))
            return true;
         return false; 
         } 
      catch (IOException ex) {
         return false;
         }
      }

   private static File firstFile(File dir) {
      if (dir == null) return null;
      File[] files = dir.listFiles(new java.io.FileFilter() {
                      @Override
                      public boolean accept(File file) {return file.isFile();}
                      });
      return (files.length > 0) ? files[0] : null;
      }

   private static File firstSubDir(File dir) {
      if (dir == null) return null;
      File[] dirs = dir.listFiles(new java.io.FileFilter() {
             @Override
             public boolean accept(File file) { return file.isDirectory();}
             });
      return (dirs.length > 0) ? dirs[0] : null;
      }

   private static boolean isSeries(File dir) {
      return isDICOM(firstFile(dir));
      }
   private static boolean isStudy(File dir) {
      return isDICOM(firstFile(firstSubDir(dir)));
      }
   private static boolean isSubject(File dir) {
      return isDICOM(firstFile(firstSubDir(firstSubDir(dir))));
      }

   private static String[] scan(File path) {
//IJ.log("Scanning: "+path.getAbsolutePath());
      String[] dirs = path.list(new java.io.FilenameFilter() {
                      @Override
                      public boolean accept(File dir, String name) {return new File(dir,name).isDirectory();}
                      });
      String[] choices = new String[2+dirs.length];
      choices[0] = "< Choose >";
      choices[1] = "< UP >";
      for(int i=0; i<dirs.length; i++) {
         File dir = new File(path, dirs[i]);
              if (isSeries(dir))  choices[i+2] = "Series:  "+dirs[i];
         else if (isStudy(dir))   choices[i+2] = "Study:   "+dirs[i];
         else if (isSubject(dir)) choices[i+2] = "Subject: "+dirs[i];
                             else choices[i+2] = "?: "+dirs[i];
         }
      return choices;
      }

   private static TreeMap<String,String> parseDICOMFile(File file, int s) {
      return parseDICOM(IJ.openImage(file.getAbsolutePath()),s);
      }
   public static TreeMap<String,String> parseDICOM(ImagePlus imp, int s) {
      TreeMap<String,String> map = new TreeMap<String,String>();
      for(String line : imp.getStack().getSliceLabel(s).split("\n",2)[1].split("\n")) {
         String[] kv = line.split(" ",2);
         if (kv[0].equals("7FE0,0010"))
            break;
         map.put(kv[0],kv[1]);
         }
      return map;
      }

   private void updateFields(File dir) {
      TreeMap<String,String> map = null;
      boolean series  = false;
      boolean study   = false;
      boolean subject = false;
           if (series  = isSeries(dir))  map = parseDICOMFile(firstFile(dir),1);
      else if (study   = isStudy(dir))   map = parseDICOMFile(firstFile(firstSubDir(dir)),1);
      else if (subject = isSubject(dir)) map = parseDICOMFile(firstFile(firstSubDir(firstSubDir(dir))),1);
                                    else map = new TreeMap<String,String>();
      info.setText(padR("Subject",13)    +padL(map.getOrDefault("0010,0020",": ").split(": ",2)[1].trim(),10)+"  "
                                         +padR(map.getOrDefault("0010,0010",": ").split(": ",2)[1].trim(),27)+"\n"
                  +padR("Study",13)       +padL(map.getOrDefault("0020,0010",": ").split(": ",2)[1].trim(),10)+"  "
                                         +padR(map.getOrDefault("0008,0020",": ").split(": ",2)[1].trim(),27)+"\n"
          +(series?padR("Series",13)+padL(map.getOrDefault("0020,0011",": ").split(": ",2)[1].trim(),10)+"  "
                                         +padR(map.getOrDefault("0008,103E",": ").split(": ",2)[1].trim(),27):" "));
      }

   public boolean dialogItemChanged(GenericDialog gd, AWTEvent e) {
      if (IJ.isMacOSX()) IJ.wait(50);
      if (e == null) return false;
      String selected = which.getSelectedItem();

      if (diff.getSelectedIndex()>0) {
         var.setText("");
         tag.setText("");
         }
      else if (e.getSource().equals(var) && !var.getText().equals("") && tag.getText().equals("")) {
         tag.setText(tags.get(var.getText()));
         }

      File dir = new File(where.getText());
      File subdir = null;
           if (which.getSelectedIndex() == 0) subdir = dir;
      else if (which.getSelectedIndex() == 1) subdir = dir.getParentFile();
                                         else subdir = new File(dir,selected.split(": *",2)[1]);
      updateFields(subdir);
      if (isSeries(subdir)) {
         ((Checkbox)gd.getCheckboxes().get(0)).setEnabled(true);
         if (gd.getNextBoolean()) {
            ((Checkbox)gd.getCheckboxes().get(0)).setState(false);
            ImagePlus imp = IJ.openImage(firstFile(subdir).getAbsolutePath());
            IJ.run(imp, "Show Info...", "");
            IJ.log(getTagLine(imp,1,"0018,0050"));
            IJ.log(getTagLine(imp,1,"0018,0080"));
            IJ.log(getTagLine(imp,1,"0018,0081"));
            IJ.log(getTagLine(imp,1,"0018,0088"));
            IJ.log(getTagLine(imp,1,"0028,1050"));
            IJ.log(getTagLine(imp,1,"0028,1051"));

            IJ.log(getTagLine(imp,1,"0018,0093"));
            IJ.log(getTagLine(imp,1,"0018,0094"));
            IJ.log(getTagLine(imp,1,"0018,1312"));
            IJ.log(getTagLine(imp,1,"0028,0010"));
            IJ.log(getTagLine(imp,1,"0028,0011"));
            }
         Button[] bs = gd.getButtons();
         bs[0].setLabel("Load Series");
         bs[0].repaint();
         return true;
         }
      else
         ((Checkbox)gd.getCheckboxes().get(0)).setEnabled(false);

      where.setText(subdir.getAbsolutePath());
      which.removeAll();
      for(String str : scan(subdir))
         which.add(str);
      return false;
      }

//////////////////////////////////////////////////////////////////////////////////
   public static ImagePlus openDICOMShort(File dfs) {
      return openDICOM(dfs, "16-bit fixed");
      }
   public static ImagePlus openDICOMShort(File[] dfs) {
      return openDICOM(dfs, "16-bit fixed");
      }
   public static ImagePlus openDICOMFloat(File dfs) {
      return openDICOM(dfs,"Float");
      }
   public static ImagePlus openDICOMFloat(File[] dfs) {
      return openDICOM(dfs,"Float");
      }
   public static ImagePlus openDICOMScaled(File dfs) {
      return openDICOM(dfs,"Scaled");
      }
   public static ImagePlus openDICOMScaled(File[] dfs) {
      return openDICOM(dfs,"Scaled");
      }

   public static ImagePlus openDICOM(File f, String format) {
      if (f.isDirectory()) return openDICOM(f.listFiles(),format);
                      else return openDICOM(new File[]{f},format);
      }
   public static ImagePlus openDICOM(File[] images, String format) {
      boolean return16bit = format.equals("16-bit fixed");
      boolean openAsFloat = !format.equals("16-bit broken");
      boolean scaleIt     = format.equals("Scaled");
      Double m = Double.NaN, b = Double.NaN;

      boolean SAVED_ignoreRescaleSlope = Prefs.ignoreRescaleSlope; Prefs.ignoreRescaleSlope = true;
      boolean SAVED_openDicomsAsFloat  = Prefs.openDicomsAsFloat;  Prefs.openDicomsAsFloat  = openAsFloat;

      ImagePlus[] imps = new ImagePlus[images.length];
      for(int i=0; i<images.length; i++) {
         IJ.showStatus("Reading DICOM Files");
         IJ.showProgress((float)i/images.length);
         imps[i] = (ImagePlus)IJ.runPlugIn("ij.plugin.DICOM", images[i].getAbsolutePath());
         if (return16bit) {
            float[][] fa = imps[i].getProcessor().getFloatArray();
            IJ.run(imps[i],"16-bit","");
            imps[i].getProcessor().setFloatArray(fa);
            }
         if (i==0)
             try { m = getDouble(imps[i],1,"0028,1053"); b = getDouble(imps[i],1,"0028,1052"); }
             catch(Throwable e) { m = 1.0; b = 0.0; }
         }
      ImagePlus imp = (new Concatenator()).concatenate(imps,true);
      if (scaleIt) {
         IJ.run(imp, "Multiply...", "value="+m+" stack");
         IJ.run(imp, "Add...",      "value="+b+" stack");
         }
      imp.setDimensions(imp.getNChannels(),imp.getNFrames(),1);
      imp.changes = false;
      IJ.showProgress(0.0);

      Prefs.ignoreRescaleSlope = SAVED_ignoreRescaleSlope;
      Prefs.openDicomsAsFloat  = SAVED_openDicomsAsFloat;

      return imp;
      }


   public static ImagePlus openDICOMDiff(File dfs, String format, String reqdirstr) {
      boolean return16bit = format.equals("16-bit fixed");
      boolean openAsFloat = !format.equals("16-bit broken");

      int reqdir = -1;
      switch(reqdirstr) {
      case "X": reqdir = 3; break;
      case "Y": reqdir = 4; break;
      case "Z": reqdir = 5; break;
      default:  reqdir = 15;
      }

      //boolean SAVED_ignoreRescaleSlope = Prefs.ignoreRescaleSlope; Prefs.ignoreRescaleSlope = true;
      boolean SAVED_openDicomsAsFloat  = Prefs.openDicomsAsFloat;  Prefs.openDicomsAsFloat  = openAsFloat;

      File[] images = dfs.listFiles();
      Vector<ImagePlus> imps = new Vector<ImagePlus>();
      for(int i=0; i<images.length; i++) {
         IJ.showStatus("Reading DICOM Files");
         IJ.showProgress((float)i/images.length);
         ImagePlus imp = (ImagePlus)IJ.runPlugIn("ij.plugin.DICOM", images[i].getAbsolutePath());

         try { int dir = getInt(imp,1,"0043,1030"); if (!((dir == 14) || (dir == reqdir))) continue;}
         catch(Throwable e) { }

         if (return16bit) {
            float[][] fa = imp.getProcessor().getFloatArray();
            IJ.run(imp,"16-bit","");
            imp.getProcessor().setFloatArray(fa);
            }

         imps.add(imp);
         }
      ImagePlus imp = (new Concatenator()).concatenate(imps.toArray(new ImagePlus[0]),true);
      imp.setDimensions(imp.getNChannels(),imp.getNFrames(),1);
      imp.changes = false;
      IJ.showProgress(0.0);

      //Prefs.ignoreRescaleSlope = SAVED_ignoreRescaleSlope;
      Prefs.openDicomsAsFloat  = SAVED_openDicomsAsFloat;

      return imp;
      }

//////////////////////////////////////////////////////////////////////////////////
   public static ImagePlus volumeDICOM(ImagePlus imp) {
      int ns = imp.getNSlices();
      Map<Double,Integer> sMap = new TreeMap<Double,Integer>();
      double[] ss = new double[ns];
      ImageStack is = imp.getStack();
      for(int s=0; s<ns; s++) {
         ss[s] = -getDouble(imp,s+1,"0020,1041");
         sMap.put(ss[s],0);
         //is.setSliceLabel(var+" = "+(vs[s])+"\n"+is.getSliceLabel(s+1).split("\n",2)[1],s+1);
         }
      { int i=0; for(Double key : sMap.keySet()) sMap.put(key,i++); }

      Map<Integer,Integer> tmpMap = new TreeMap<Integer,Integer>();
      for(int s=0; s<ns; s++) 
         tmpMap.put(sMap.get(ss[s]), s+1);
      String list = "";
      for(Integer key : tmpMap.keySet())
         list += ","+tmpMap.get(key);
      ImagePlus tmpimp = SubHyperstackMaker.makeSubhyperstack(imp,"1",list.substring(1),"1");
      tmpimp.setDimensions(1,sMap.size(),1);
      //tmpimp.setOpenAsHyperStack(true);
      //tmpimp.setTitle(var+" - "+imp.getTitle());

      return tmpimp;
      }
//////////////////////////////////////////////////////////////////////////////////
   public static ImagePlus arrayDICOM(ImagePlus imp, String tag, String var) {
      int ns = imp.getNSlices();
      double[] vs = new double[ns];
      Map<Double,Integer> sMap = new TreeMap<Double,Integer>();
      double[] ss = new double[ns];
      Map<Double,Integer> vMap = new TreeMap<Double,Integer>();
      ImageStack is = imp.getStack();
      for(int s=0; s<ns; s++) {
         ss[s] = -getDouble(imp,s+1,"0020,1041");
         sMap.put(ss[s],0);
         try { vs[s] = getDouble(imp,s+1,tag); }
         catch(Throwable e) { IJ.error("tag \""+tag+"\" does not exist"); return null; }
         vMap.put(vs[s],0);
         is.setSliceLabel(var+" = "+(vs[s])+"\n"+is.getSliceLabel(s+1).split("\n",2)[1],s+1);
         }
      { int i=0; for(Double key : vMap.keySet()) vMap.put(key,i++); }
      { int i=0; for(Double key : sMap.keySet()) sMap.put(key,i++); }

      Map<Integer,Integer> tmpMap = new TreeMap<Integer,Integer>();
      for(int s=0; s<ns; s++) 
         tmpMap.put(vMap.get(vs[s])*sMap.size() + sMap.get(ss[s]), s+1);
      String list = "";
      for(Integer key : tmpMap.keySet())
         list += ","+tmpMap.get(key);
      ImagePlus tmpimp = SubHyperstackMaker.makeSubhyperstack(imp,"1",list.substring(1),"1");
      tmpimp.setDimensions(1,sMap.size(),vMap.size());
      tmpimp.setOpenAsHyperStack(true);
      tmpimp.setTitle(var+" - "+imp.getTitle());

      return tmpimp;
      }

//////////////////////////////////////////////////////////////////////////////////
   public static ImagePlus[] diffusionDICOM(ImagePlus imp) throws Exception {

      int ns = imp.getNSlices();
      Map<Integer,Integer> inMap = new TreeMap<Integer,Integer>();
      Map<Double,Integer> bvMap = new TreeMap<Double,Integer>();
      Map<Double,Integer> slMap = new TreeMap<Double,Integer>();
      int[]    in = new int[ns];
      double[] bv = new double[ns];
      double[] sl = new double[ns];

      String vendor  = null;
      String release = null;
      try {
         vendor  = "UNKNOWN"; vendor  = getTagValue(imp,1,"0008,0070");
         release = "UNKNOWN"; release = getTagValue(imp,1,"0018,1020");
         for(int s=0; s<ns; s++) {
            IJ.showStatus("Sorting Diffusion");
            IJ.showProgress((float)s/ns);
            sl[s] = -getDouble(imp,s+1,"0020,1041");
            slMap.put(sl[s],0);
   
                 if (vendor.equals("GE MEDICAL SYSTEMS")) {
               bv[s] = Math.max(getArray(imp,s+1,"0043,1039")[0] -  1000000000, 0);
               bvMap.put(bv[s],getInt(imp,s+1,"0018,0083"));

               try { in[s] = getInt(imp,s+1,"0043,1030"); }
               catch(Throwable e) { in[s] = (bv[s] == 0) ? 14 : 15; }
               }
            else if (vendor.equals("SIEMENS")) {
               String dd = getTagValue(imp,s+1,"0019,100D");
               if (!dd.equals("ISOTROPIC")) 
                  throw new IllegalArgumentException("Siemens diffusion direction 0019,100D '"+dd+"' not handled!");

               bv[s] = getDouble(imp,s+1,"0019,100C");
               bvMap.put(bv[s],getInt(imp,s+1,"0018,0083"));

               if (bv[s] < 1) in[s] = 14;
                         else in[s] = 15;
               }
            else {
               bv[s] = getDouble(imp,s+1,"0018,9087");
               bvMap.put(bv[s],getInt(imp,s+1,"0018,0083"));
            
               if (bv[s] < 1) in[s] = 14;
                         else in[s] = 15;
               }
            }
         }
      catch(Exception e) {
         IJ.error("The following error is probably either due to dataset is not DICOM\nor a DICOM tag unique to unsupported vendor '"+vendor+"' or release '"+release+"'");
         throw e;
         }

      String[] types = new String[]{"Trace","R/L","A/P","S/I"};
      inMap.put(15, 0); // Trace
      inMap.put( 3, 1); // X R/L
      inMap.put( 4, 2); // Y A/P
      inMap.put( 5, 3); // Z S/I
      inMap.put(14,-1); // b=0


      { int i=0; for(Double key : bvMap.keySet()) bvMap.put(key,i++); }
      { int i=0; for(Double key : slMap.keySet()) slMap.put(key,i++); }
      ImagePlus[] impArr = new ImagePlus[inMap.size()-1];

//for(Double key : bvMap.keySet()) IJ.log("b="+key);
//for(Double key : slMap.keySet()) IJ.log("s="+key);

      String b0list = "";
      {
         Map<Integer,Integer> tmpMap = new TreeMap<Integer,Integer>();
         for(int s=0; s<ns; s++) 
            if (in[s] == 14)
               tmpMap.put(bvMap.get(bv[s])*slMap.size() + slMap.get(sl[s]), s+1);
         for(Integer tmpKey : tmpMap.keySet())
            b0list += ","+tmpMap.get(tmpKey);
      }
      if (b0list.equals(""))
         IJ.error("There were no b=0 images.");

      for(Integer inKey : inMap.keySet()) {
         if (inKey == 14)
            continue;
         Integer n = inMap.get(inKey);
         Map<Integer,Integer> tmpMap = new TreeMap<Integer,Integer>();
         for(int s=0; s<ns; s++) {
            if (in[s] != inKey)
               continue;
            tmpMap.put(bvMap.get(bv[s])*slMap.size() + slMap.get(sl[s]), s+1);
            }
         if (tmpMap.size() == 0)
            continue;
         String list = b0list;
         for(Integer tmpKey : tmpMap.keySet())
            list += ","+tmpMap.get(tmpKey);
         String[] rev = list.substring(1).split(",");
         impArr[n] = SubHyperstackMaker.makeSubhyperstack(imp,"1",list.substring(1),"1");
         ImageStack is = impArr[n].getStack();
         for(int s=0; s<impArr[n].getNSlices(); s++) {
            int rs = Integer.parseInt(rev[s]) - 1;
            double b = bv[rs];
            if (b<0) b=0;
            is.setSliceLabel("b = "+((int)b)+"\n"+is.getSliceLabel(s+1).split("\n",2)[1],s+1);
            }
	     impArr[n].setDimensions(1,slMap.size(),bvMap.size());
	     impArr[n].setOpenAsHyperStack(true);
         impArr[n].setTitle(types[n]+" - "+imp.getTitle());
         }

      IJ.showStatus("Sorting Diffusion Done");
      IJ.showProgress(0.0);
      return impArr;
      }

   public static double[]  getArray(ImagePlus imp, int s, String tag) {
      String[] vs = getTagValue(imp, s, tag).split("\\\\");
      double[] ia = new double[vs.length];
      for(int i=0; i<vs.length; i++)
         ia[i] = Double.parseDouble(vs[i]);
      return ia;
      }
   public static double getDouble(ImagePlus imp, int s, String tag) {
      return Double.parseDouble(getTagValue(imp, s, tag));
      }
   public static int getInt(ImagePlus imp, int s, String tag) {
      return Integer.parseInt(getTagValue(imp, s, tag));
      }
   public static String getTagValue(ImagePlus imp, int s, String tag) {
      for(String line : imp.getStack().getSliceLabel(s).split("\n"))
         if (line.startsWith(tag))
            return line.split(": ",2)[1].trim();
      return null;
      }
   public static String getTagLine(ImagePlus imp, int s, String tag) {
      for(String line : imp.getStack().getSliceLabel(s).split("\n"))
         if (line.startsWith(tag))
            return line;
      return null;
      }

//////////////////////////////////////////////////////////////////////////////////
}
